#!/usr/bin/python
from classGraph import Graph, NodeMap, TuplesCover, Clause
from toolLog import Log
import pdb, re, random, numpy, subprocess, datetime, os, os.path, copy, itertools, sys, pickle, glob

########################################################
################# READ PARAMETERS
def readParameters(fp=None):
    tf_dict = {"True":True, "False":False, "true":True, "false": False}

    ### pickle: -1 disabled, 0 store, 1 load
    ### log: - stdout, + auto filename, * both, other filename
    ### normalize comp: flip comparison edges A>B to A<B
    ### try simply: try finding simpler clauses when remapping 
    ### try V: which shapes to try 1 /\, 2 \/, 3 -- positive indices: even if simple edge has enough support
    org_parameters = {'verbosity': 3, 'logfile': "+", 'unlink': True, 'pickle': -1,
                      'redescribe': True, 'list_all': True, "print_supp": False,
                      'limit_unit_paths': 0, 'limit_abs_paths':-(10**5),
                      'limit_unit_instances': 0, 'limit_abs_instances': 10**5,
                      'limit_length': 0, 'limit_time': 60,
                      'limit_turns': 4, 'limit_relsupp': 0, 'limit_ratio': 2, 'last_paths':False,
                      'try_extend': 1, 'try_init': 5, 'alt_longer': 2,
                      'singleton_map_check': False, 'limit_clause': 0,
                      'min_examples':3, 'limit_contrib': 3, 'id_exporg': None, 'id_exp': "TMP",
                      'results_rep': "../results/", "data_rep":"../data/",
                      "query_nodes": ['pA', 'pZ'], "normalize_comp": True,
                      "try_V": [-1, -2], "try_simply": 1,
                      'limit_paths_itemsets': 1000, 'limit_graphs_itemsets': 1000,
                      'limit_multiplicity': 3, 'remap_reds':True,
                      'list_okey': Clause.basic_okey
                      }

    # 'limit_paths_itemsets': 1000, 'limit_graphs_itemsets': 1000,
    # 'limit_unit_instances': 0, 'limit_abs_instances': 10**5,

    if fp is not None:
        read_parameters = {}
        for line in fp:
            if re.match("^ *#", line) is not None:
                continue
            parts = line.split()
            if len(parts) > 1:
                if parts[1] in tf_dict:
                    read_parameters[parts[0]] = tf_dict[parts[1]]
                else:
                    try:
                        read_parameters[parts[0]] = int(parts[1])
                    except ValueError:
                        try:
                            read_parameters[parts[0]] = float(parts[1])
                        except ValueError:
                            tmp = re.sub("~", os.path.expanduser("~"), parts[1])
                            read_parameters[parts[0]] = tmp
        org_parameters.update(read_parameters)

    if org_parameters.has_key('okey'):
        org_parameters['list_okey'] = []
        for part in org_parameters['okey'].split(','):
            tmp = part.split(':')
            if len(tmp) > 1:
                org_parameters['list_okey'].append((tmp[0], float(tmp[1])))
            else:
                org_parameters['list_okey'].append((tmp[0], -1))
                
    list_accept = []
    for (k,v) in org_parameters.items():
        fact = 1
        tmp = re.match("^(?P<Dir>(min)|(max))_(?P<Crit>[\w]+)$", k)
        if tmp is not None:
            if tmp.group("Dir") == "max":
                fact = -1
            list_accept.append((tmp.group("Crit"), v, fact))
    if len(list_accept) > 0:
        org_parameters['list_accept'] = list_accept

    if not org_parameters.has_key('min_support'): org_parameters['min_support'] = 0
    if not org_parameters.has_key('min_examples'): org_parameters['min_examples'] = 0
    if not org_parameters.has_key('min_precision'): org_parameters['min_precision'] = 0
    if not org_parameters.has_key('min_recall'): org_parameters['min_recall'] = 0
    if not org_parameters.has_key('min_jaccard'): org_parameters['min_jaccard'] = 0
    if not org_parameters.has_key('max_ratio'): org_parameters['max_ratio'] = org_parameters['limit_ratio']

    return org_parameters

def openLog(parameters, log=None):
    if log is not None:
        parameters['log'] = log
    else:
        if parameters['logfile'] == "*":
            parameters['log'] = Log(parameters['verbosity'], "-")
            parameters['logfile'] = "+"

        tmp_logfile = parameters['logfile']
        if parameters['logfile'] == "+":
            interm = ""
            if parameters['id_exporg'] is not None:
                interm = parameters['id_exporg'] + "_"
            tmp_logfile = parameters['results_rep'] + interm + parameters['id_exp'] + ".log"
        parameters['log'] = Log(parameters['verbosity'], tmp_logfile)

########################################################
################# INITIALIZE CLAUSE FROM EDGE
def initClauses(pred, val, data, parameters):
    clauses = []
    inv_att = data.filteredPredTypes(pred)
    neighbors_dict = data.graph.neighborsSub(-1, inv_att[Graph.EG].keys())
    edges = data.graph.matchingElements(Graph.EG, [(pred, val)])
    limit_instances = parameters['limit_abs_instances']
    if parameters['limit_unit_instances'] != 0 and parameters['limit_unit_instances']*len(edges) < limit_instances:
        limit_instances = parameters['limit_unit_instances']*len(edges)
    
    filtered_edges = [e for e in edges if len(neighbors_dict[e[0]])*len(neighbors_dict[e[1]]) > 0]
    has_base = False
    if (parameters['limit_contrib'] < 0 or len(filtered_edges) >= parameters['limit_contrib']) and ( not parameters.has_key('min_examples') or len(filtered_edges) >= parameters['min_examples']) :
        has_base = True
        key = "%s_%s%s" % (parameters['id_exp'], pred, val)
        pv = "%s=%s" % (pred, val)
        clause = Clause(key, [('pA', pv ,'pZ')], data.graph, parameters['query_nodes'], set(), data.toClausePreds())
        if clause.support() is not None and ( \
            (parameters['limit_contrib'] < 0 or len(clause.support()) >= parameters['limit_contrib']) and ( not parameters.has_key('min_examples') or len(clause.support()) >= parameters['min_examples'])) :
            clauses.append(clause)

    if len(parameters["try_V"]) > 0 and (parameters["try_V"][0] > 0 or not has_base):

        exclus_n = [{}, {}]
        for edge in edges:
            exclus_n[0].setdefault(edge[0], []).append(edge[1])
            exclus_n[1].setdefault(edge[1], []).append(edge[0])

        nb_pairs = [0,0,0]
        for endi in [0,1]:
            nb_pairs[endi] = 2*sum([len(v)*(len(v)-1)/2 for v in exclus_n[endi].values()])

        nb_pairs[2] = sum([len(v)*len(exclus_n[1].get(k,[])) for k,v in exclus_n[0].items()])

        # nb_conn = [0,0]
        # for endi in [0,1]:
        #     nb_conn[endi] = len([e for e in edges if len(neighbors_dict[e[endi]]) > 0])

        # if nb_conn[0]+nb_conn[1] > 0 and nb_conn[0]*nb_conn[1] == 0:
        #     zero_side = 0
        #     if nb_conn[0] > 0: zero_side = 1

        #     dict_conn = {}
        #     for edge in edges:
        #         dict_conn.setdefault(edge[zero_side], []).append(edge[1-zero_side])
        #     nb_pairs = 2*sum([len(v)*(len(v)-1)/2 for v in dict_conn.values()])

        parameters["log"].printL(4, "Trying V shapes, %s=%s: %s..." % (pred, val, str(nb_pairs)))
        for tmp_side in parameters["try_V"]:
            zero_side = abs(tmp_side)-1
            if ( limit_instances == 0 or nb_pairs[zero_side] <= limit_instances ) and \
                ( parameters['limit_contrib'] < 0 or nb_pairs[zero_side] >= parameters['limit_contrib']) and \
                ( not parameters.has_key('min_examples') or nb_pairs[zero_side] >= parameters['min_examples']):

                key = "%s_%dV%s%s" % (parameters['id_exp'], zero_side, pred, val)
                pv = "%s=%s" % (pred, val)
                if zero_side == 0:
                    atts = [('p1', pv ,'pA'), ('p1', pv ,'pZ')]
                elif zero_side == 1:
                    atts = [('pA', pv ,'p1'), ('pZ', pv ,'p1')]
                else:
                    atts = [('pA', pv ,'p1'), ('p1', pv ,'pZ')]

                clause = Clause(key, atts, data.graph, parameters['query_nodes']) 
                clauses.append(clause)


    return clauses


########################################################
################# PREPARE DICT OF PARAMETERS FOR ROUND
def prepareParameters(org_parameters, to_expand, collect_clauses, data, log=None):

    patt = collect_clauses[to_expand['clause']]
    edges = patt.support()
    if edges is None:
        return [], None
    else:
        edges = list(edges)
    nb_edges = len(edges)
    if to_expand['turn'] > org_parameters['limit_turns'] or \
       ( org_parameters['limit_contrib'] > 0 and len(edges) < org_parameters['limit_contrib']) or \
       ( org_parameters.has_key('min_examples') and len(edges) < org_parameters['min_examples']) :
        return [], None

    filter_types = patt.getTypesPred()
    if len(filter_types) == 0:
        Warning("No predicates exculded, that's odd...")
        filter_types = patt.getTypesPred()

    parameters = copy.deepcopy(org_parameters)
    parameters['query_nodes'] = ['pA', 'pZ']
    parameters['id_exp'] = patt.getId()
    parameters['id_orgexp'] = org_parameters["id_exp"]

    openLog(parameters, log)

    parameters['inv_att'] = data.filteredPredTypes(filter_types)
        
    Graph.log = parameters['log']

    parameters['nb_edges'] = nb_edges
    
    if to_expand['other_side'] is not None:
        parameters['min_jaccard'] = patt.getCrit('jaccard')
        if not parameters.has_key('list_accept'):
            parameters['list_accept']= []
        parameters['list_accept'].append(("jaccard", parameters['min_jaccard'], 1))
        try:
            parameters['clause_oneback'] = collect_clauses[to_expand['other_side']]
        except IndexError:
            Warning("No such clause")

    if parameters['min_support'] < 0: parameters['min_support'] *= -nb_edges

    if parameters['limit_contrib'] >= 0:
        parameters['min_examples'] = max([parameters['min_examples'], parameters['limit_contrib']])
    parameters['thres_support'] = max([parameters['min_examples'], parameters['min_support'],
                                       parameters['min_jaccard']*nb_edges, parameters['min_recall']*nb_edges])
    if parameters.has_key('max_nb_nodes') and ( parameters['limit_length'] == 0 or parameters['limit_length'] > parameters['max_nb_nodes']-1):
        parameters['limit_length'] = parameters['max_nb_nodes']-1
    return edges, parameters
    
########################################################
################# MAIN
def searchClause(graph, edges, parameters):

    if parameters['pickle'] <= 0:
        tic = datetime.datetime.now()
        parameters["log"].printL(2, "Mine paths...")
        itemsets, paths_reached, path_covered_blues, nodes_labels_dict = minePaths(graph, edges, parameters)

        # save_p = (itemsets, paths_reached, path_covered_blues, nodes_labels_dict)
        # fp = open("../data/debugPaths.pickle", "w")
        # pickle.dump(save_p, fp)
        # fp.close()

        # fp = open("../data/debugPaths.pickle")
        # save_p = pickle.load(fp)
        # fp.close()
        # (itemsets, paths_reached, path_covered_blues, nodes_labels_dict) = save_p

        toc = datetime.datetime.now()
        parameters["log"].printL(2, "Filter...")
        itemsets, singletons = filterItemsetsSingletons(itemsets, paths_reached, nodes_labels_dict, parameters)
        parameters["log"].printL(2, "Hash...")
        hashes, itemsets_ids = hashRolesPaths(itemsets, paths_reached, path_covered_blues, singletons, graph, parameters)
        parameters["log"].printL(2, "Enumerate graphs...")
        graphs_lists, cons_keys, connecting = enumerateConstrainedGraphs(hashes, itemsets, nodes_labels_dict, parameters, itemsets_ids)
        
        basename = parameters['results_rep']+("%s_graphs" % parameters['id_exp'])
        parameters["log"].printL(2, "Mine graphs...")
        #itemsets_data, covered_ids = mineGraphsItemsetsCP(connecting, graphs_lists, parameters, basename)
        itemsets_data, covered_ids = mineGraphsItemsetsEclat(connecting, graphs_lists, parameters, basename)
        roots, itemsets_list = generateInclusionTree(itemsets_data)

        if parameters['pickle'] == 0:
            save_p = (itemsets_list, roots, cons_keys, path_covered_blues)
            fp = open("../data/debugClauses.pickle", "w")
            pickle.dump(save_p, fp)
            fp.close()
    else:

        tic = datetime.datetime.now()
        toc = datetime.datetime.now()
        fp = open("../data/debugClauses.pickle")
        save_p = pickle.load(fp)
        fp.close()
        (itemsets_list, roots, cons_keys, path_covered_blues) = save_p


    parameters["log"].printL(2, "Construct clauses...")
    clauses = []
    deja_vus = {}

    parameters['count_ex'] = 0
    for root in roots:
        constructClausesTree(0, root, itemsets_list, clauses, graph, edges, cons_keys, parameters, deja_vus)
    parameters['count_cl'] = len(clauses)

    #     if parameters['pickle'] == 0:
    #         save_p = (clauses, path_covered_blues)
    #         fp = open("../data/debugClauses.pickle", "w")
    #         pickle.dump(save_p, fp)
    #         fp.close()

    # else:

    #     tic = datetime.datetime.now()
    #     toc = datetime.datetime.now()
    #     fp = open("../data/debugClauses.pickle")
    #     save_p = pickle.load(fp)
    #     fp.close()
    #     (clauses, path_covered_blues) = save_p

    parameters["log"].printL(2, "Select clauses...")
    selected, covered_tuples = selectClauses(clauses, graph, path_covered_blues, edges, parameters)

    if parameters['list_all']:
        filename = parameters['results_rep']+("%s_all.bmg" % parameters['id_exp'])
        writeClausesList(clauses, filename)
    
    tac = datetime.datetime.now()
    parameters["log"].printL(2, "THE END for %s, support %f. Covers %d of %d reached out of %d blue tuples, and %d red tuples, with %d clauses... Spent %s mining paths, %s combining." % (parameters['id_exp'], parameters['thres_support'], len(set(edges) & covered_tuples), len(path_covered_blues), len(edges), len(covered_tuples - set(edges)), len(selected), (toc-tic), (tac-toc)))

    # del itemsets
    # del paths_reached
    # del hashes
    # del itemsets_list
    # del itemsets_data
    # del graphs_lists
    details = {'clauses': clauses, 'covered_tuples': covered_tuples, 'path_covered_blues': path_covered_blues, 'edges': edges, 'tic': tic, 'toc': toc, 'tac':tac, 'selected': selected}
    return  selected, details 

def searchPaths(graph, edges, parameters):
    supps = dict([(e,ei) for (ei, e) in enumerate(sorted(edges))])
    supps_outs = set()
    supps_ins = set()
    pathst = []
    if parameters['pickle'] <= 0:
        tic = datetime.datetime.now()
        parameters["log"].printL(2, "Mine paths...")
        itemsets, paths_reached, path_covered_blues, nodes_labels_dict = minePaths(graph, edges, parameters)

        for path in itemsets:
            mapped_nodes, details = graph.mapClause(path['path'].getConstraintsList(), parameters['query_nodes'], None, True, parameters['limit_abs_instances'])
            cover = TuplesCover(parameters['query_nodes'], mapped_nodes, edges)
            supps_outs |= cover.redTuples()
            supps_ins |= cover.blueTuples()
            pathst.append((path['path'], cover))

    for (ei, e) in enumerate(sorted(supps_outs)):
        supps[e] = -(ei+1)
    paths = []
    for pt in pathst:
        try:
            paths.append((pt[0], [supps[e] for e in pt[1].blueTuples()]+[supps[e] for e in pt[1].redTuples()]))
        except KeyError as ek:
            pdb.set_trace()
            print ek

        
    return paths, supps

#################### FINDING PATHS COVERING THE EXAMPLES
##########################################################
def minePaths(graph, edges, parameters):
    nodes_labels_dict = {}
    itemsets = []
    paths_reached = []
    paths_list = []
    path_covered_blues = set()
    destinations = {}
    thres_support_lp = []
    covered_steps = []
    covert_change_sets = []

    neighbors_dict = graph.neighborsSub(-1, parameters['inv_att'][Graph.EG].keys())
    filtered_edges = [e for e in edges if len(neighbors_dict[e[0]])*len(neighbors_dict[e[1]]) > 0]
    # filtered_edges = edges
    #################### Prepare list of starting nodes with destinations
    counts_ext = {}
    for edgei in range(len(filtered_edges)):
        if not destinations.has_key(filtered_edges[edgei][0]):
            paths_list.append([filtered_edges[edgei][0]])
            destinations[filtered_edges[edgei][0]] = {}
            counts_ext[filtered_edges[edgei][0]] = 1
        destinations[filtered_edges[edgei][0]][filtered_edges[edgei][1]] = edgei

    #################### Look for connecting paths, increasing length
    plength = 0
    while not parameters['last_paths'] and len(paths_list) > 0:
        plength += 1
        if parameters['limit_length'] > 0 and plength == parameters['limit_length']:
            parameters['last_paths'] = True
        parameters["log"].printL(2, "EXTENDING PATHS...")
        (paths_list, tmp_reach, counts_ext) = joinLists(paths_list, neighbors_dict, plength, destinations, parameters, counts_ext)
        reached_pairs = set([(tr[-1], tr[0]) for tr in tmp_reach])
        parameters["log"].printL(3, "Path length:%d, Nb pairs:%d reached out of %d, Nb paths:%d reached, %d others" % (plength, len(reached_pairs), len(edges), len(tmp_reach), len(paths_list)))
        (paths_transactions, cons_keys, nodes_labels) = enumerateConstrainedPaths(graph, tmp_reach, parameters['inv_att'])

        # thres_support_lp.append(int(parameters['min_support']*len(reached_pairs | path_covered_blues)))
        # if thres_support_lp[-1] < parameters['min_examples']:
        #     thres_support_lp[-1] = parameters['min_examples']

        # thres_support_lp.append(parameters['thres_support'])
        thres_support_lp.append(max(parameters['limit_relsupp']*len(reached_pairs), parameters['thres_support']))

        basename = parameters['results_rep']+("%s-%s_paths" % (parameters['id_exp'], plength))

        parameters["log"].printL(2, "MINING CLAUSES...")
        if parameters["normalize_comp"]:
            normalizeCompConstraints(cons_keys)
        # itemsets_tmp, covered_pairs = minePathsItemsetsCP(plength, paths_transactions, cons_keys, tmp_reach, edges, thres_support_lp[-1], parameters['limit_time'], parameters, basename)
        itemsets_tmp, covered_pairs = minePathsItemsetsEclat(plength, paths_transactions, cons_keys, tmp_reach, edges, thres_support_lp[-1], parameters['limit_time'], parameters, basename)
        nb_new = len(itemsets_tmp)
        paths_reached.append(tmp_reach)
        nodes_labels.reverse()
        tmp_nodes = dict([(n, pi) for (pi, n) in enumerate(nodes_labels)])
        tmp_nodes['pZ'] = -1
        nodes_labels_dict.update(tmp_nodes)
        itemsets.extend(itemsets_tmp)
                
        #################### Update coverage info
        covered_steps.append(covered_pairs - path_covered_blues)
        path_covered_blues |= covered_pairs 
        parameters["log"].printL(3, "Selected %d(+%d) clauses, covering %d/%d pairs (%d/%d new)" % (len(itemsets), nb_new , len(path_covered_blues), len(edges), len(covered_steps[-1]), len(covered_pairs)))

        covert_change_sets.append(list(covered_steps[-1]))
        parameters["log"].printL(3, "Cover changed: %s" % ([len(e) for e in covert_change_sets]))

        if len(covert_change_sets) > parameters['try_init'] and sum([len(cp) for cp in covert_change_sets]) == 0:
        ### Not reached anything in first 2*try_extend steps, stop
            tmp_list = []
            parameters["log"].printL(3, "Giving up, really not reaching...")
        elif sum([len(cp) for cp in covert_change_sets]) > 0 and \
                 ((len(covert_change_sets) > parameters['try_extend'] and sum([len(covert_change_sets[-sti]) for sti in range(1,parameters['try_extend']+1)]) == 0) \
                  or (len(covered_steps) > parameters['try_extend'] and sum([len(covered_steps[-sti]) for sti in range(1,parameters['try_extend']+1)]) == 0)):
        ### No new pairs reach during last try_extend steps, stop
            tmp_list = []
            parameters["log"].printL(3, "No more trying...")
        elif parameters['alt_longer'] > 0 and len(covert_change_sets) >= parameters['alt_longer']:
        ### else, remove starting nodes the reached all their targets alt_longer steps ago
            parameters["log"].printL(3, "Discard all reached %d steps ago..." % parameters['alt_longer'])
            for edge in covert_change_sets[-parameters['alt_longer']]:                    
                if destinations.has_key(edge[0]) and destinations[edge[0]].has_key(edge[1]):
                    del destinations[edge[0]][edge[1]]
                ### previously was equivalent to:
                ## destinations[edge[0]] = {}
            tmp_list = [p for p in paths_list+tmp_reach if len(destinations[p[-1]]) > 0]
        else:
            tmp_list = paths_list + tmp_reach
            ### previously was equivalent to:
            ## tmp_list = paths_list
        paths_list = tmp_list


    del paths_list
    del destinations
    del covered_steps
    del covert_change_sets
    return itemsets, paths_reached, path_covered_blues, nodes_labels_dict


def joinLists(org_list, extensions_dict, cycle_check=0, destinations=None, parameters=None, counts_ext=None):
    exclude_ext = set()
    exclude_nodes = set()
    ext_list = []
    reached_list = []
    estimate_nb_paths = 0

    limit_abs = parameters['limit_abs_paths']
    if limit_abs < 0:
        limit_abs *= -parameters["nb_edges"] 

    if counts_ext is not None:
        if parameters['limit_unit_paths'] > 0:
            exclude_nodes = set([node for node in counts_ext.keys() if counts_ext[node]*len(extensions_dict[node]) > parameters['limit_unit_paths']])
            parameters["log"].printL(5, "Path extension, exclude %d nodes" % len(exclude_nodes))
        if limit_abs > 0:
            estimate_nb_paths = sum([count*len(extensions_dict[node]) for node, count in counts_ext.items()])
            parameters["log"].printL(5, "Path extension, estimate %d paths" % estimate_nb_paths)
            if estimate_nb_paths > limit_abs:
                parameters['last_paths'] = True
                
        if parameters['last_paths']:
            counts_ext = None
        else:
            counts_ext = dict([(n, 0) for n in extensions_dict.keys()])

    ### IF COUNT AND ESTIMATE IS OVER THRESHOLD DON'T EVEN TRY
    for ol in org_list:
        if ol[0] in exclude_nodes:
            continue
        
        if limit_abs > 0 and (len(ext_list) + len(reached_list)) > limit_abs:
            if parameters['last_paths']:
                ext_list = []
                reached_list = []
                break
            else:
                parameters['last_paths'] = True
                ext_list = []

        if cycle_check > 0:
            exclude_ext = set(ol[:cycle_check])
        extensions = set(extensions_dict.get(ol[0], set()) - exclude_ext)
        destinations_o = set(destinations[ol[-1]].keys())
        if counts_ext is not None:
            for ext in extensions:
                counts_ext[ext] += 1
                
        reached_list.extend([[ext]+ol for ext in (extensions & destinations_o)])
        if not parameters['last_paths']:
            ext_list.extend([[ext]+ol for ext in (extensions - destinations_o)])
        
    return (ext_list, reached_list, counts_ext)


def normalizeCompConstraints(constraints_keys):
    for cid, const in enumerate(constraints_keys):
        if Graph.isEG(const[0]) and const[1][1] == Graph.comp_labels[Graph.COMP_GT]:
            constraints_keys[cid] = [(const[0][1], const[0][0]), (const[1][0], Graph.comp_labels[Graph.COMP_LT])]
            

def constraintsFromIds(constraints_ids, constraints_keys):
    constraints_tmp = Graph()
    for constraint in constraints_ids:
        const = list(constraints_keys[constraint])
        constraints_tmp.parseElem(Graph.typeE(const[0]), const[0], const[1][0], const[1][1])
    return constraints_tmp

#################### CONSTRUCT ITEMSETS FOR PATHS
##########################################################
def enumerateConstrainedPaths(graph, paths, inv_att, parameters=None, preff=""):

    if parameters is None or not parameters.has_key('query_nodes'):
        query_nodes = ['pA', 'pZ']
    else:
        query_nodes = parameters['query_nodes']
    nodes_labels = []
    labels_enum = []
    cons_keys = []
    paths_transactions = []
    for pi, path in enumerate(paths):
        paths_transactions.append([])
        if pi == 0:
            if type(preff) == list and len(preff) == len(graph):
                nodes_labels = list(preff)
            else:
            ### Paths are built in reverse order !
                nodes_labels = ['p%s%d' % (preff, elemi) for elemi in range(len(path)-1,-1,-1)]
                nodes_labels[0] = query_nodes[-1]
                nodes_labels[-1] = query_nodes[0]

            elemi = -1
            for elemi in range(len(path)-1): ## all paths have same length
                labels_enum.extend([nodes_labels[elemi], (nodes_labels[elemi], nodes_labels[elemi+1]), (nodes_labels[elemi+1], nodes_labels[elemi])])
                for elemj in range(elemi+1):
                    labels_enum.append((nodes_labels[elemj], nodes_labels[elemi+1]))
            labels_enum.append(nodes_labels[elemi+1])
        cp = 0
        for constraint_list in graph.getConstrainedPath(path, inv_att):
            for constraint in constraint_list:
                cons_key = (labels_enum[cp], constraint)
                if not cons_key in cons_keys:
                    cons_keys.append(cons_key)
                paths_transactions[-1].append(cons_keys.index(cons_key))
            cp += 1
    return (paths_transactions, cons_keys, nodes_labels)

#################### CONSTRUCT ITEMSETS FOR GRAPHS
##########################################################
def enumerateConstrainedGraphs(hashes, itemsets, nodes_labels_dict, parameters, itemsets_ids):    

    attributes = {}
    attributes[None] = {Graph.ND: {}, Graph.EG: {}}
    for pair, roles in hashes.items():
        curr_hash = hashes[pair]
        attributes[pair] = {Graph.ND: {}, Graph.EG: {}}
        for pid, card in roles['pids'].items():
            itemset = itemsets[pid]
            for qid in range(card):
                for elem, atts in itemset['path'].getNodes():
                    if elem not in parameters['query_nodes']:
                        nid = curr_hash['nodes_keys'][curr_hash['reversed'][(pid, nodes_labels_dict[elem], qid)]]
                    else:
                        nid = elem
                    
                    attributes[pair][Graph.ND].setdefault(nid, set()).update(atts.items())
                    attributes[None][Graph.ND].setdefault(nid, set()).update(atts.items())

                for elem, atts in itemset['path'].getEdges():
                    eid = list(elem)
                    for endi in [0, 1]:
                        if elem[endi] not in parameters['query_nodes']:
                            eid[endi] = curr_hash['nodes_keys'][curr_hash['reversed'][(pid, nodes_labels_dict[elem[endi]], qid)]]
                    eid = tuple(eid)
                    attributes[pair][Graph.EG].setdefault(eid, set()).update(atts.items())
                    attributes[None][Graph.EG].setdefault(eid, set()).update(atts.items())

    cons_keys = []
    connecting = set()
    for tE in [Graph.ND, Graph.EG]:
        for elem, atts in attributes[None][tE].items():
            for att in atts:
                if tE == Graph.EG and not Graph.isComp(att):
                   connecting.add(len(cons_keys)) 
                cons_keys.append((elem, att))

    graphs_lists = {}
    for pair, pat in attributes.items():
        if pair is None:
            continue
        graphs_lists[pair] = []
        for tE in [Graph.ND, Graph.EG]:
            for elem, atts in attributes[pair][tE].items():
                for att in atts:
                    graphs_lists[pair].append(cons_keys.index((elem, att)))

    return graphs_lists, cons_keys, connecting

#################### MINE FREQUENT PATHS ITEMSETS
##########################################################
def minePathsItemsetsCP(plength, paths_transactions, cons_keys, paths, edges, thres_support, limit_time, parameters, basename="paths"):
    itemsets = []
    covered_pairs = set()
    cut_time = None

    if len(paths_transactions) == 0:
        return itemsets, covered_pairs
    ### Connectivity items constraints sets
    connected_sets = {}
    for itemid, cons_key in enumerate(cons_keys):
        if Graph.isEG(cons_key[0]) and not Graph.isComp(cons_key[1]):
            connected_sets.setdefault((min(cons_key[0]), max(cons_key[0])), set()).add(itemid)
    connection_list = connected_sets.values()

    ### Write data for CP
    fp = open(basename+".cp_transactions", "w")
    for set_c in connection_list:
        fp.write('> '+' '.join(map(str, set_c))+'\n')

    trans_dict = []
    trans_counts = []
    tpcs = {}
    for (tid, path_c) in enumerate(paths_transactions):
        ind = edges.index((paths[tid][-1], paths[tid][0]))
        tpc = tuple(path_c)
        if tpcs.has_key(tpc):
            trans_dict[tpcs[tpc]].setdefault(ind, []).append(tid)
            trans_counts[tpcs[tpc]] += 1
        else:
            tpcs[tpc] = len(trans_dict)
            trans_dict.append({ind: [tid]})
            trans_counts.append(1)

    edges_tp = []
    inline_c = 0
    for tpc, pos in tpcs.items(): 
        if parameters['limit_multiplicity'] <= 0 \
               or trans_counts[pos] < len(trans_dict[pos])*parameters['limit_multiplicity']:
            for ind in trans_dict[pos].keys():
                inline_c += 1
                fp.write("%s %d\n" % (' '.join(map(str, tpc)), ind))
                edges_tp.append((pos, ind))
    fp.close()

    if inline_c == 0:
        return itemsets, covered_pairs
    
    ### Run CP
    call_params = ["./mine_paths_alt.sh", basename, "%d" % thres_support, "%d" % (parameters['limit_paths_itemsets']*parameters['limit_multiplicity']), "%d" %  parameters['limit_time']]
    subprocess.call(call_params)

    ### Parse CP output
    fp = open(basename+".cp_itemsets")
    gid = 0
    dropm = 0
    log_line = "---"
    for line in fp:
        gid +=1
        if cut_time == None:
            log_line = line
            cut_time = line[0] == '1'
            continue
        if gid-1-dropm > parameters['limit_paths_itemsets']:
            break
        
        partsU = re.match('^(?P<itemset>[0-9 ]*)\t(?P<path_freq>\d+):(?P<paths_ids>[0-9 ]*)\t(?P<pair_freq>\d+):(?P<pairs_ids>[0-9 ]*)$', line)
        if partsU != None and int(partsU.group('pair_freq')) >= thres_support:
            tmp_itemset = {'constraints_keys': map(int, partsU.group('itemset').split()),
                           'plength': plength,
                           'blue_pairs': [edges[int(x)] for x in partsU.group('pairs_ids').split()],
                           'path_freq': int(partsU.group('path_freq')),
                           'pair_freq': int(partsU.group('pair_freq'))}         
            tmp = []
            paths_ids = []
            for tk in map(int, partsU.group('paths_ids').split()):
                pos, ind = edges_tp[tk]
                paths_ids.extend(trans_dict[pos][ind])
                tmp.append(ind)

            if parameters['limit_multiplicity'] > 0 and len(tmp) > parameters['limit_multiplicity']*len(set(tmp)):
                dropm += 1
                continue

            tmp_itemset['paths_ids'] = paths_ids
            tmp_itemset['order_key'] = (tmp_itemset['pair_freq'] == tmp_itemset['path_freq'], tmp_itemset['pair_freq'], tmp_itemset['plength'])
            tmp_itemset['path'] = constraintsFromIds(tmp_itemset['constraints_keys'], cons_keys)
            itemsets.append(tmp_itemset)
            covered_pairs.update(tmp_itemset['blue_pairs'])
    fp.close()

    parameters["log"].printL(3, "Mining %s\n\t%d transactions with min_supp %d and timeout %ds, parsed %d and dropped %d high multiplicity...\n\t[%s]" % (basename, len(paths_transactions), thres_support, limit_time, gid-1, dropm, log_line.strip()))

    itemsets.sort(key= lambda x: (x['pair_freq'], len(x['constraints_keys'])), reverse= True)
    ### REMOVE non-closed on pairs (not only on instances)
    # i = 0
    # unf_itnb = len(itemsets)
    # while i < len(itemsets):
    #     j = i-1
    #     while j >= 0 and itemsets[i]['pair_freq'] == itemsets[j]['pair_freq']:
    #         if set(itemsets[i]['blue_pairs']) == set(itemsets[j]['blue_pairs']) and set(itemsets[i]['constraints_keys']) < set(itemsets[j]['constraints_keys']):
    #             itemsets.pop(i)
    #             j=-1
    #             i-=1
    #         else:
    #             j-=1
    #     i+=1
    ### TODO CHECK DO WE WANT THAT            
    # del itemsets[500:]

    ### remove CP in/out files
    if parameters['unlink']:
        for fn in glob.glob(basename+".*"):
            os.unlink(fn)
    return itemsets, covered_pairs

def minePathsItemsetsEclat(plength, paths_transactions, cons_keys, paths, edges, thres_support, limit_time, parameters, basename="paths"):
    itemsets = []
    covered_pairs = set()
    cut_time = None

    if len(paths_transactions) == 0:
        return itemsets, covered_pairs

    ### Connectivity items constraints sets
    connected_sets = {}
    for itemid, cons_key in enumerate(cons_keys):
        if Graph.isEG(cons_key[0]) and not Graph.isComp(cons_key[1]):
            connected_sets.setdefault((min(cons_key[0]), max(cons_key[0])), set()).add(itemid)
    connection_list = connected_sets.values()

    trans_dict = []
    trans_counts = []
    tpcs = {}
    for (tid, path_c) in enumerate(paths_transactions):
        ind = edges.index((paths[tid][-1], paths[tid][0]))
        tpc = tuple(path_c)
        if tpcs.has_key(tpc):
            trans_dict[tpcs[tpc]].setdefault(ind, []).append(tid)
            trans_counts[tpcs[tpc]] += 1
        else:
            tpcs[tpc] = len(trans_dict)
            trans_dict.append({ind: [tid]})
            trans_counts.append(1)

    ### Write data for MINER
    fp = open(basename+".transactions", "w")

    edges_tp = []
    inline_c = 0
    for tpc, pos in tpcs.items(): 
        if parameters['limit_multiplicity'] <= 0 \
               or trans_counts[pos] < len(trans_dict[pos])*parameters['limit_multiplicity']:
            for ind in trans_dict[pos].keys():
                inline_c += 1
                fp.write(" ".join(map(str, tpc)) + "\n")
                edges_tp.append((pos, ind))
    fp.close()

    if inline_c == 0:
        return itemsets, covered_pairs
    
    ### Run CP
    call_params = ["./mine_paths_nv.sh", basename, "%d" % thres_support, "%d" % (parameters['limit_paths_itemsets']*parameters['limit_multiplicity']), "%d" %  parameters['limit_time']]
    subprocess.call(call_params)

    ### Parse CP output
    itemsets_data = {}
    count_fr = {}
    ffr = open(basename+".itemsets")
    ftr = open(basename+".support")
    li = 0
    for line in ffr:        
        lsupp = ftr.readline()
        itemset = set(map(int, line.split()))
        linked = True
        i = 0
        while linked and i < len(connection_list):
            linked &= len(connection_list[i].intersection(itemset)) > 0
            i += 1

        if linked:
            paths_ids = [i-1 for i in map(int, lsupp.split())]
            pairs_ids = set([edges_tp[i][1] for i in paths_ids])
            pair_freq = len(pairs_ids)
            path_freq = len(paths_ids)
                
            if pair_freq >= thres_support:
                if not count_fr.has_key(pair_freq):
                    count_fr[pair_freq]=1
                else:
                    count_fr[pair_freq]+=1

                if not itemsets_data.has_key(len(itemset)):
                    itemsets_data[len(itemset)] = []
                itemsets_data[len(itemset)].append({"parents": set(), "children": set(), "ancestors": set(), "constraints_keys": itemset,
                                                    "path_freq": path_freq, "pair_freq": pair_freq,
                                                    "paths_ids": paths_ids, "pairs_ids": pairs_ids, "leftover": set(pairs_ids)})
    ftr.close()
    ffr.close()

    ### remove MINER files
    if parameters['unlink']:
        for fn in glob.glob(basename+".*"):
            os.unlink(fn)

    filterItemsetsData(itemsets_data, count_fr, parameters['limit_paths_itemsets'])
    alls, roots, leaves, drops = getRootsEtc(itemsets_data)

    alls_list = sorted(alls)
    map_nd = dict([(v,k) for (k,v) in enumerate(alls_list)])
    roots_list = []
    for l,p in alls_list:

        current = itemsets_data[l][p]

        blue_pairs =set()
        blue_paths = []
        for tk in current['paths_ids']:
            pos, ind = edges_tp[tk]
            blue_paths.extend(trans_dict[pos][ind])
            blue_pairs.add(edges[ind])

        tmp_itemset = current
        tmp_itemset["plength"] = plength
        tmp_itemset["paths_ids"] = blue_paths
        tmp_itemset["blue_pairs"] = blue_pairs
        tmp_itemset['order_key'] = (tmp_itemset['pair_freq'] == tmp_itemset['path_freq'], tmp_itemset['pair_freq'], tmp_itemset['plength'])
        tmp_itemset['path'] = constraintsFromIds(tmp_itemset['constraints_keys'], cons_keys)

        tmp_itemset["children"] = [map_nd[c] for c in tmp_itemset["children"] & alls]
        prt_nb = len(tmp_itemset["parents"] & alls)
        if prt_nb == 0:
            roots_list.append(len(itemsets))
            tmp_itemset["parent"] = None
            tmp_itemset["diff"] = tmp_itemset['constraints_keys']
        else:
            tmp = sorted(tmp_itemset["parents"] & alls)
            tmp_itemset["parent"] = map_nd[tmp[0]]
            tmp_itemset["diff"] = tmp_itemset['constraints_keys'] - itemsets[tmp_itemset["parent"]]['constraints_keys']
            for pr in tmp[1:]:
                try:
                    itemsets[map_nd[pr]]["children"].remove(len(itemsets))
                except ValueError:
                    print "No such child"
                    pdb.set_trace()

        del tmp_itemset["parents"]
        del tmp_itemset["leftover"]
        del tmp_itemset["ancestors"] 

        itemsets.append(tmp_itemset)
        covered_pairs.update(blue_pairs)

    parameters["log"].printL(3, "Mining %s\n\t%d transactions with min_supp %d, found %d and kept %d..." % (basename, len(paths_transactions), thres_support, sum(count_fr.values()), len(itemsets)))

    itemsets.sort(key= lambda x: (x['pair_freq'], len(x['constraints_keys'])), reverse= True)

    return itemsets, covered_pairs


def filterItemsetsData(itemsets_data, count_fr, max_i=2000):
    mv, agg = -1, 0
    counts = sorted(count_fr.items())
    while agg < max_i and len(counts) > 0:
        mv, tmp = counts.pop()
        agg += tmp

    if len(counts) > 0:
        for clen in itemsets_data.keys():
            itemsets_data[clen] = [it for it in itemsets_data[clen] if it["pair_freq"] >= mv]

def getRootsEtc(itemsets_data):
    addParentLinks(itemsets_data)

    roots, leaves, drops = set(), set(), set()
    for clen, its in itemsets_data.items():
        for cp, it in enumerate(its):
            if len(it['parents']) == 0:
                roots.add((clen, cp))
            elif len(it['children']) == 0:
                leaves.add((clen, cp))
            elif len(it["leftover"]) > 0:
                drops.add((clen, cp))

    ilengths = sorted(itemsets_data.keys(), reverse=True)
    alls = set(leaves | drops | roots)
    for clen in ilengths:
        for cp, it in enumerate(itemsets_data[clen]):
            if (clen, cp) in alls and len(it["parents"]) > 0 and len(it["parents"] & alls) == 0:
                sel = sorted(it["parents"], key=lambda x: len(itemsets_data[x[0]][x[1]]["children"]))[-1]
                alls.add(sel)

    return alls, roots, leaves, drops

def addParentLinks(itemsets_data):
    # tic = datetime.datetime.now()
    ilengths = sorted(itemsets_data.keys())
    cci = 1
    while cci < len(ilengths):
        cur_len = ilengths[cci]
        cur_pos = 0
        while cur_pos < len(itemsets_data[cur_len]):
            ci = cci-1
            while ci >= 0:
                comp_len = ilengths[ci]
                cands = set([(comp_len,i) for i in range(len(itemsets_data[comp_len]))]) - itemsets_data[cur_len][cur_pos]['ancestors']
                for dum, comp_pos in cands:
                    if itemsets_data[cur_len][cur_pos]['constraints_keys'] > itemsets_data[comp_len][comp_pos]['constraints_keys']:
                        itemsets_data[cur_len][cur_pos]['parents'].add( (comp_len, comp_pos) )
                        itemsets_data[cur_len][cur_pos]['ancestors'].update(itemsets_data[comp_len][comp_pos]['parents'])
                        itemsets_data[cur_len][cur_pos]['ancestors'].update(itemsets_data[comp_len][comp_pos]['ancestors'])
                        itemsets_data[comp_len][comp_pos]['children'].add( (cur_len, cur_pos) )
                        itemsets_data[comp_len][comp_pos]['leftover'] -= itemsets_data[cur_len][cur_pos]['pairs_ids']
                ci -=1
            cur_pos += 1
        cci +=1
    # print "DONE ", datetime.datetime.now() - tic
 
#################### MINE FREQUENT GRAPHS ITEMSETS
##########################################################
def mineGraphsItemsetsCP(connecting, graphs_lists, parameters, basename="graphs"):
    ### thres_supp = max(parameters['min_examples'], parameters['min_support']*len(graphs_lists))
#    thres_support = max(parameters['limit_relsupp']*len(graphs_lists), parameters['thres_support'])
    thres_support = parameters['thres_support']
    covered_ids = set()
    itemsets_data = {}
    cut_time = None

    if len(connecting) == 0:
        return itemsets_data, covered_ids

    pairs_keys = sorted(graphs_lists.keys())
    
    ### Write data for CP
    fp = open(basename+"cp_transactions", "w")
    fp.write('> '+' '.join(map(str, connecting))+'\n')
    for tid, pair in enumerate(pairs_keys):
        fp.write("%s %d\n" % (' '.join(map(str, graphs_lists[pair])), tid))
    fp.close()

    ### Run CP
    call_params = ["./mine_paths.sh", basename, "%d" % thres_support, "%d" % (parameters['limit_paths_itemsets']*parameters['limit_multiplicity']), "%d" %  parameters['limit_time']]
    subprocess.call(call_params)

    ### Parse CP output
    fp = open(basename+".cp_itemsets")
    gid=0
    log_line = "---"
    for line in fp:
        gid +=1
        if cut_time == None:
            log_line = line
            cut_time = line[0] == '1'
            continue
        if gid-1 > parameters['limit_graphs_itemsets']:
            break

        partsU = re.match('^(?P<itemset>[0-9 ]*)\t(?P<path_freq>\d+):(?P<paths_ids>[0-9 ]*)\t(?P<pair_freq>\d+):(?P<pairs_ids>[0-9 ]*)$', line)
        if partsU != None:
            tmp_itemset = {'constraints_keys': set(map(int, partsU.group('itemset').split())),
                           'paths_ids': map(int, partsU.group('paths_ids').split()),
                           'parent': None,
                           'id': gid,
                           'children': [],
                           'path_freq': int(partsU.group('path_freq'))}
            tmp_itemset['blue_pairs'] = [pairs_keys[pk] for pk in tmp_itemset['paths_ids']]
            itemsets_data.setdefault(len(tmp_itemset['constraints_keys']), []).append(tmp_itemset)
            covered_ids.update(tmp_itemset['blue_pairs'])
    fp.close()
    parameters["log"].printL(3, "Mining %s\n\t%d transactions with min_supp %d and timeout %ds, parsed %d...\n\t[%s]" % (basename, len(pairs_keys), thres_support, parameters['limit_time'], gid-1, log_line.strip()))

    ### remove CP in/out files
    if parameters['unlink']:
        for fn in glob.glob(basename+".*"):
            os.unlink(fn)
    return itemsets_data, covered_ids


def mineGraphsItemsetsEclat(connecting, graphs_lists, parameters, basename="graphs"):
    ### thres_supp = max(parameters['min_examples'], parameters['min_support']*len(graphs_lists))
    thres_support = parameters['thres_support']
    covered_ids = set()
    itemsets_data = {}
    cut_time = None

    if len(connecting) == 0:
        return itemsets_data, covered_ids

    pairs_keys = sorted(graphs_lists.keys())
    
    ### Write data for CP
    fp = open(basename+".transactions", "w")
    for pair in pairs_keys:
        fp.write(' '.join(map(str, graphs_lists[pair])) + "\n")
    fp.close()

    ### Run CP
    call_params = ["./mine_graphs.sh", basename, "%d" % thres_support, "%d" % (parameters['limit_paths_itemsets']*parameters['limit_multiplicity']), "%d" %  parameters['limit_time']]
    subprocess.call(call_params)

    ### Parse CP output
    fp = open(basename+".cp_itemsets")
    gid=0
    log_line = "---"
    for line in fp:
        gid +=1
        if cut_time == None:
            log_line = line
            cut_time = line[0] == '1'
            continue
        if gid-1 > parameters['limit_graphs_itemsets']:
            break

        partsU = re.match('^(?P<itemset>[0-9 ]*)\t(?P<graph_freq>\d+):(?P<graphs_ids>[0-9 ]*)$', line)
        if partsU != None:
            tmp_itemset = {'constraints_keys': set(map(int, partsU.group('itemset').split())),
                           'graphs_ids': [int(u)-1 for u in partsU.group('graphs_ids').split()],
                           'parent': None, 'id': gid, "children": [],
                           'graph_freq': int(partsU.group('graph_freq'))}
            tmp_itemset['blue_pairs'] = [pairs_keys[pk-1] for pk in tmp_itemset['graphs_ids']]

            tmp_itemset.update({'pairs_ids': set(tmp_itemset['blue_pairs']),
                                'leftover': set(tmp_itemset['blue_pairs']),
                                'children': set(), 'parents': set(), 'ancestors': set()})

            itemsets_data.setdefault(len(tmp_itemset['constraints_keys']), []).append(tmp_itemset)
            covered_ids.update(tmp_itemset['blue_pairs'])
    fp.close()
    parameters["log"].printL(3, "Mining %s\n\t%d transactions with min_supp %d and timeout %ds, parsed %d...\n\t[%s]" % (basename, len(pairs_keys), thres_support, parameters['limit_time'], gid-1, log_line.strip()))

    ### remove CP in/out files
    if parameters['unlink']:
        for fn in glob.glob(basename+".*"):
            os.unlink(fn)
    return itemsets_data, covered_ids

def queryConnects(constraints, queries):
    edges = [e for e in constraints.keys() if Graph.isEG(e)]
    nodeset = set()
    node_to_part = {}
    part_to_node = []
    for edge in edges:
        nodeset.update(edge)
    if not (queries[0] in nodeset and queries[1] in nodeset):
        return False

    for edge in edges:
        if not node_to_part.has_key(edge[0]) and not node_to_part.has_key(edge[1]):
            node_to_part[edge[0]] = len(part_to_node)
            node_to_part[edge[1]] = len(part_to_node)
            part_to_node.append(set(edge))
        elif not node_to_part.has_key(edge[0]):
            node_to_part[edge[0]] = node_to_part[edge[1]]
            part_to_node[node_to_part[edge[0]]].add(edge[0])
        elif not node_to_part.has_key(edge[1]):
            node_to_part[edge[1]] = node_to_part[edge[0]]
            part_to_node[node_to_part[edge[1]]].add(edge[1])
        elif node_to_part[edge[1]] != node_to_part[edge[0]]:
            part_to_node[node_to_part[edge[1]]] |= part_to_node[node_to_part[edge[0]]]
            for node in part_to_node[node_to_part[edge[0]]]:
                node_to_part[node] = node_to_part[edge[1]]
            part_to_node[node_to_part[edge[0]]] = set()
    return node_to_part.get(queries[0], -1) == node_to_part.get(queries[1], -2)
    
#################### CONSTRUCTION OF CLAUSES FROM ITEMSETS TREE
##########################################################
def constructClausesTree(level, cid, itemsets_list, clauses, graph, edges, cons_keys, parameters, deja_vus=None, basis=None, constraints_out=None, prev_mapped_nodes=None):
    level +=1
    parameters['count_ex'] +=1
    max_score = parameters['limit_abs_instances']
    if parameters['limit_unit_instances'] != 0 and parameters['limit_unit_instances']*len(itemsets_list[cid]['blue_pairs']) < max_score:
        max_score = parameters['limit_unit_instances']*len(itemsets_list[cid]['blue_pairs'])
    
    if deja_vus is None:
        deja_vus = {}
    
    constraints_process = {}
    if constraints_out is not None:
        for elem, atts in constraints_out:
            if constraints_process.has_key(elem):
                constraints_process[elem].update(atts)
            else:
                constraints_process[elem] = dict(atts)

    for item in itemsets_list[cid]['diff']:
        (elem, att) = cons_keys[item]
        if constraints_process.has_key(elem):
            constraints_process[elem].update({att[0]: att[1]})
        else:
            constraints_process[elem] = {att[0]: att[1]}
    constraints_list = constraints_process.items()

    if prev_mapped_nodes is None and not queryConnects(constraints_process, parameters['query_nodes']):
        parameters["log"].printL(2, "%s JUMP EXPANDING %s (%d)" % ( (level*" ")+" |__", cid, len(itemsets_list[cid]['children'])))
        constraints_in = basis
        reprocess = constraints_list
        for child in itemsets_list[cid]['children']:
            constructClausesTree(level, child, itemsets_list, clauses, graph, edges, cons_keys, parameters, deja_vus, constraints_in, reprocess, prev_mapped_nodes)
        return
    
    parameters["log"].printL(4, "Mapping %d %d\n" % ( cid, len(itemsets_list[cid]['blue_pairs'])))
    mapped_nodes, details = graph.mapClause(constraints_list, parameters['query_nodes'], prev_mapped_nodes, True, max_score)
    
    if details['status'] == 0:
        #### SOMETHING WENT BADLY WRONG
        ## pdb.set_trace()
        parameters["log"].printL(1, "SHORT ON MAPPING")
        if basis is not None:
            parameters["log"].printL(7, "\n".join(basis.listBMG()))
        parameters["log"].printL(7, "\n\tPLUS\n%s" % str(constraints_list))
    else:
        
        expand = True
        if len(constraints_list) == len(details['reprocess']):
        ## No change happened, just forward....
            constraints_in = basis
            reprocess = constraints_list
            
        else:
        ## Some change happened
            if basis is None:
                constraints_in = Graph()
            else:
                constraints_in = basis.copy()

            reprocess = []
            for elem, atts in constraints_list:
                if elem in details['reprocess']:
                    reprocess.append((elem, atts))
                else:
                    for att in atts.items():
                        constraints_in.parseElem(Graph.typeE(elem), elem, att[0], att[1])

            itemset_in = []
            for elem, atts in constraints_in.getConstraintsList():
                for att in atts.items():
                    itemset_in.append(cons_keys.index((elem, att)))
            itemset_in_key = '.'.join(map(str, sorted(itemset_in)))
            
            if deja_vus.has_key(itemset_in_key):
                deja_vus[itemset_in_key]['cids'].append(cid)
                expand = deja_vus[itemset_in_key]['expand']

            else:
                #### CHECK query nodes are connected
                expand = True
                ### If query nodes are not connected, we cannot bound support... 

                if len(details['heads_present']) == len(parameters["query_nodes"]) and \
                   len(details['head_fragments']) == 1:


                    #### GOT AN UNSEEN CONNECTED CLAUSE...

                    ### filter disconnected nodes
                    ### keep only head fragment
                    ### exactly one head fragment if we end up here...
                    pnodes_map = mapped_nodes.projectFragment(details['head_fragments'].pop())
                    pconstraints = constraints_in.copy()
                    pconstraints.project(pnodes_map.nodesSet())

                    clause_data = Clause("%s.%d" % (parameters['id_exp'], cid), pconstraints, pnodes_map, parameters['query_nodes'], edges)
 
                    if clause_data.hasSingletonMap():
                        if parameters['singleton_map_check']:
                            parameters["log"].printL(2, "%s#### %s SINGLETON" % ( (level*" ")+"'-"  , clause_data.dispShort()))
                            return
                        else:
                            clause_data.labelSingletonMaps()


                    ### SHOULD WE ACCEPT THE NEW CLAUSE, SHOULD WE EXPAND IT?
                    accept, log_check = clause_data.checkAccept(parameters)
                    parameters["log"].printL(2, "%s %s Accept:%s" % ( (level*" ")+""  , clause_data.dispShort(), accept))
                    parameters["log"].printL(3, "%s [detail:%s]" % ( ((level+2)*" ")+""  , log_check))
                    if accept:
                        clauses.append(clause_data)                
                    expand, log_check = clause_data.checkExpand(parameters)
                    parameters["log"].printL(2, "%s %s Expand:%s" % ( (level*" ")+""  , clause_data.dispShort(), expand))
                    parameters["log"].printL(3, "%s [detail:%s]" % ( ((level+2)*" ")+""  , log_check))


                ### ADD the new clause wether connected or not
                deja_vus[itemset_in_key] = {'cids': [cid], 'expand': expand}

        if expand:
            parameters["log"].printL(2, "%s EXPANDING %s (%d)" % ( (level*" ")+" |__", cid, len(itemsets_list[cid]['children'])))
            for child in itemsets_list[cid]['children']:
                constructClausesTree(level, child, itemsets_list, clauses, graph, edges, cons_keys, parameters, deja_vus, constraints_in, reprocess, mapped_nodes)
            
def filterItemsetsSingletons(itemsets, paths_reached, nodes_labels_dict, parameters):
    filtered = []
    singletons = {}
    nb_singletons = 0
    nodes_labels_invert = [None for i in range(len(nodes_labels_dict))]
    for label, pos in nodes_labels_dict.items():
        nodes_labels_invert[pos] = label
    i=0
    while len(itemsets) > 0:
        itemset = itemsets.pop()
        i+=1
        maps_versions = {}
        if itemset['plength'] > 1 and itemset['path_freq'] % itemset['pair_freq'] == 0:
            nb_times = itemset['path_freq'] / itemset['pair_freq']
            #### WARNING, WARNING, WARNING, PATHS ARE INVERTED!
            nodes_to = zip(*[paths_reached[itemset['plength']-1][path_id][1:-1]  for path_id in itemset['paths_ids']])
            for nid, node_to in enumerate(nodes_to):
                if len(set(node_to)) == nb_times:
                    versions = sorted(set(node_to))
                    tmp_versions = dict([(version, []) for version in versions])
                    for path_id in itemset['paths_ids']:
                        pair = (paths_reached[itemset['plength']-1][path_id][-1], paths_reached[itemset['plength']-1][path_id][0])
                        tmp_versions[paths_reached[itemset['plength']-1][path_id][nid+1]].append((pair, path_id))
                    same_length = True
                    base_len = itemset['pair_freq']
                    comp = 0
                    while comp < len(versions) and same_length:
                        same_length = len(tmp_versions[versions[comp]]) == base_len
                        comp +=1
                    if same_length:
                        maps_versions[nid] = tmp_versions

        if len(maps_versions) == 0:
            filtered.append(itemset)

        elif parameters['singleton_map_check']:
            parameters["log"].printL(4, "Singleton detected, dropping...")
            nb_singletons += 1
            
        elif len(maps_versions) == 1:
            nb_singletons += 1
            nid, map_versions = maps_versions.popitem()
            parameters["log"].printL(4, "One singleton detected gives %s out..." % ([(len(filtered) + itmi) for itmi in range(len(map_versions))]))
            for version, map_version in map_versions.items():
                tmp_itemset = {'constraints_keys': list(itemset['constraints_keys'])+[None],
                               'plength': itemset['plength'],
                               'sing': "unique%d" % i,
                               'paths_ids': [v[1] for v in map_version],
                               'blue_pairs': list(itemset['blue_pairs'])}
                tmp_itemset['path_freq']= len(tmp_itemset['paths_ids'])
                tmp_itemset['pair_freq']= len(tmp_itemset['blue_pairs'])         
                tmp_itemset['order_key']= (tmp_itemset['pair_freq'] == tmp_itemset['path_freq'], tmp_itemset['pair_freq'], tmp_itemset['plength'])
                tmp_itemset['path']= itemset['path'].copy()
                tmp_itemset['path'].parseElem(Graph.ND, nodes_labels_invert[itemset['plength']-(nid+1)], Graph.singleton_label, version)
                singletons.setdefault(version, []).append(len(filtered))
                filtered.append(tmp_itemset)
        else:
            nb_singletons += 1
            if nb_times == 1:
                parameters["log"].printL(4, "%d unique singletons detected gives out %s..." % (len(maps_versions), [len(filtered)]))
                tmp_itemset = {'constraints_keys': list(itemset['constraints_keys'])+[None]*nb_times,
                               'plength': itemset['plength'],
                               'sing': "delta%d" % i,
                               'paths_ids': list(itemset['paths_ids']),
                               'blue_pairs': list(itemset['blue_pairs'])}
                tmp_itemset['path_freq']= len(tmp_itemset['paths_ids'])
                tmp_itemset['pair_freq']= len(tmp_itemset['blue_pairs'])         
                tmp_itemset['order_key']= (tmp_itemset['pair_freq'] == tmp_itemset['path_freq'], tmp_itemset['pair_freq'], tmp_itemset['plength'])
                tmp_itemset['path']= itemset['path'].copy()

                for nid, map_versions in maps_versions.items():
                    version, map_version = map_versions.popitem()
                    tmp_itemset['path'].parseElem(Graph.ND, nodes_labels_invert[itemset['plength']-(nid+1)], Graph.singleton_label, version)
                    singletons.setdefault(version, []).append(len(filtered))
                filtered.append(tmp_itemset)                
            else:
                parameters["log"].printL(4, "%d singletons detected..." % len(maps_versions))
                # if not parameters['singleton_map_check']:
                filtered.append(itemset)
                # pdb.set_trace()
                # print maps_versions.keys()
    parameters["log"].printL(3, "Total singletons detected %d..." % nb_singletons)
    return filtered, singletons


#################### GENERATE COMBINATIONS ROLES ON DATA
##########################################################
def hashRolesPaths(itemsets, paths_reached, enter_pairs, singletons, graph, parameters):
    itemsets_ids = range(len(itemsets))
    itemsets_ids.sort(key=lambda x: itemsets[x]['order_key'], reverse=True)
    itemsets.sort(key=lambda x: x['order_key'], reverse=True)

    hashes =  dict([(p, {'signs': {}, 'reversed': {}, 'pids': {}, 'map':{}}) for p in enter_pairs])
    for pid, itemset in enumerate(itemsets):
        # if None in itemset['constraints_keys']: pdb.set_trace()
        for path_id in itemset['paths_ids']:
            path = paths_reached[itemset['plength']-1][path_id]
            pair = (path[-1], path[0])
            if pair in enter_pairs:
                if not hashes[pair]['pids'].has_key(pid):
                    hashes[pair]['pids'][pid] = 0
                qid = hashes[pair]['pids'][pid]
                hashes[pair]['reversed'][(pid, 0, qid)] = parameters['query_nodes'][0]
                hashes[pair]['reversed'][(pid, itemset['plength'], qid)] = parameters['query_nodes'][-1]
                hashes[pair]['pids'][pid] += 1
                ### enumerate in reverse order to get the from pA to pZ
                ## DEBUG if None in itemset['constraints_keys']: pdb.set_trace()
                for ii in range(1, itemset['plength']):
                    hashes[pair]['signs'].setdefault(graph.nodeId(path[-(ii+1)]), set()).add((pid, ii, qid))
                    hashes[pair]['reversed'][(pid, ii, qid)] = graph.nodeId(path[-(ii+1)])

    keys_dict = {}
    for sin in singletons.keys():
        pos = "s%s" % graph.nodeId(sin) # (len(keys_dict)+1)
        keys_dict[pos] = pos

    for pair in hashes.keys():
        nodes_labels_simple = dict([(v,k) for (k,v) in enumerate(parameters['query_nodes'])])
        nodes_labels = {}
        labels_nodes = {}
        # paths_used = set()
        for node in hashes[pair]['signs'].keys():
            hashes[pair]['signs'][node] = sorted(hashes[pair]['signs'][node])
            nodes_labels_simple[node] = hashes[pair]['signs'][node][0][:2]
    
        for elem, key_roles in hashes[pair]['signs'].items():
            if keys_dict.has_key('s%s' % elem): ### singleton node, at least somewhere...
                pos = keys_dict['s%s' % elem]
                nodes_labels[elem] = pos
                labels_nodes.setdefault(pos, []).append(elem)
            else:
                path_pos = key_roles[0]
                prec = hashes[pair]['reversed'][(path_pos[0], path_pos[1]-1, path_pos[2])]
                succ = hashes[pair]['reversed'][(path_pos[0], path_pos[1]+1, path_pos[2])]
                kr = (nodes_labels_simple[elem], nodes_labels_simple[prec], nodes_labels_simple[succ])
                try:
                    pos = keys_dict[kr]
                except KeyError:
                    pos = "p%d" % (len(keys_dict)+1)
                    keys_dict[kr] = pos
                nodes_labels[elem] = pos
                labels_nodes.setdefault(pos, []).append(elem)
                # paths_used.add(path_pos[:2])
        # if pair == ('Laboratory_or_Test_Result', 'Cell_Function'):
        #     print "####################################################"
        #     print labels_nodes

        changed = set()
        for pos, nodes in labels_nodes.items():
            if len(nodes) > 1:
                changed.update(nodes)
        qrn = set(parameters['query_nodes'])

        while len(changed) > 0:
            parameters["log"].printL(7, "Neighbors changed %s" %( changed))
            rkeys = {}
            for node in changed:
                rkeysn = set()
                for path_pos in hashes[pair]['signs'][node]:
                    prec = hashes[pair]['reversed'][(path_pos[0], path_pos[1]-1, path_pos[2])]
                    succ = hashes[pair]['reversed'][(path_pos[0], path_pos[1]+1, path_pos[2])]
                    precl = nodes_labels.get(prec, prec)
                    succl = nodes_labels.get(succ, succ)
                    rkeysn.add((path_pos[0], path_pos[1], precl, succl))
                rl = tuple(sorted(rkeysn))
                rkeys.setdefault(rl, []).append(node)

            seen = set()
            srk = sorted(rkeys.keys())
            neighbors = set()
            for rkey in srk:
                ### all nodes in batch should originally have same label, by definition
                if nodes_labels[rkeys[rkey][0]] not in seen:
                    seen.add(nodes_labels[rkeys[rkey][0]])
                else:
                    for r in rkey:
                        neighbors.update(r[2:])
                    try:
                        pos = keys_dict[rkey]
                    except KeyError:
                        pos = "p%d" % (len(keys_dict)+1)
                        keys_dict[rkey] = pos

                    parameters["log"].printL(7, "Change multiple %s %s %s" %( pos, rkeys[rkey], rkey))
                    for elem in rkeys[rkey]:
                        labels_nodes[nodes_labels[elem]].remove(elem)
                        nodes_labels[elem] = pos
                        labels_nodes.setdefault(pos, []).append(elem)

            changed  = set()
            for neigh in (neighbors - qrn):
                if len(labels_nodes[neigh]) > 1:
                    changed.update(labels_nodes[neigh])

            if len(changed) == 0:
                for lb, nodes in labels_nodes.items():
                    adjacents = []
                    if len(nodes) > 1:
                        rolest = hashes[pair]['signs'][nodes[0]]
                        for i in range(len(rolest)-1):
                            if rolest[i][0] == rolest[i+1][0] and rolest[i][1] == rolest[i+1][1]-1:
                                adjacents.append(rolest[i])
                    if len(adjacents)> 0:
                        parameters["log"].printL(7, "Change adjacent multiple %s %s" %( lb, nodes))
                        precs = set()
                        succs = set()
                        for nid, elem in enumerate(nodes):
                            for path_pos in hashes[pair]['signs'][elem]:
                                precs.add(hashes[pair]['reversed'][(path_pos[0], path_pos[1]-1, path_pos[2])])
                                succs.add(hashes[pair]['reversed'][(path_pos[0], path_pos[1]+1, path_pos[2])])
                            if nid > 0:
                                pos = lb + ".%d" % nid
                                nodes_labels[elem] = pos
                                labels_nodes.setdefault(pos, []).append(elem)
                                labels_nodes[lb].remove(elem)

                        if len(succs) > 1: changed.update(succs)
                        if len(precs) > 1: changed.update(precs)
                changed -= qrn

        parameters["log"].printL(7, "Pair hashs %s %s" %( pair, labels_nodes))
        hashes[pair]['nodes_keys'] = nodes_labels
        hashes[pair]['keys_nodes'] = labels_nodes
    
    return hashes, itemsets_ids

#################### GENERATION OF ITEMSETS INCLUSION TREE
##########################################################
def generateInclusionTree(itemsets_data):
    alls, roots, leaves, drops = getRootsEtc(itemsets_data)

    itemsets  = []
    alls_list = sorted(alls)
    map_nd = dict([(v,k) for (k,v) in enumerate(alls_list)])
    roots_list = []
    for l,p in alls_list:

        tmp_itemset = itemsets_data[l][p]
        
        tmp_itemset["children"] = [map_nd[c] for c in tmp_itemset["children"] & alls]
        prt_nb = len(tmp_itemset["parents"] & alls)
        if prt_nb == 0:
            roots_list.append(len(itemsets))
            tmp_itemset["parent"] = None
            tmp_itemset["diff"] = tmp_itemset['constraints_keys']
        else:
            tmp = sorted(tmp_itemset["parents"] & alls)
            tmp_itemset["parent"] = map_nd[tmp[0]]
            tmp_itemset["diff"] = tmp_itemset['constraints_keys'] - itemsets[tmp_itemset["parent"]]['constraints_keys']
            for pr in tmp[1:]:
                try:
                    itemsets[map_nd[pr]]["children"].remove(len(itemsets))
                except ValueError:
                    print "No such child"
                    pdb.set_trace()

        del tmp_itemset["parents"]
        del tmp_itemset["leftover"]
        del tmp_itemset["ancestors"] 

        itemsets.append(tmp_itemset)
    return roots_list, itemsets


def generateInclusionTreeOV(itemsets_data):
    ### construct inclusion tree
    for ilen, itemsets_l in itemsets_data.items():
        itemsets_l.sort(key=lambda x: sorted(x))

    itemsets_list = []
    trans_pos = {}
    ilengths = sorted(itemsets_data.keys())
    while len(ilengths) > 0:
        curr_len = ilengths.pop()
        while len(itemsets_data[curr_len]) > 0:
            cli = len(ilengths)-1
            curr_pos = len(itemsets_data[curr_len])-1
            trans_pos[(curr_len, curr_pos)] = len(itemsets_list)
            itemsets_list.append(itemsets_data[curr_len].pop(0))
            while cli >= 0 and itemsets_list[-1]['parent'] is None:
                comp_len = ilengths[cli]
                comp_pos = 0
                while comp_pos < len(itemsets_data[comp_len]) and itemsets_list[-1]['parent'] is None:
                    if itemsets_list[-1]['constraints_keys'] > \
                           itemsets_data[comp_len][comp_pos]['constraints_keys']:
                        itemsets_list[-1]['parent'] = (comp_len, comp_pos)
                        itemsets_list[-1]['diff'] = list(itemsets_list[-1]['constraints_keys'] - itemsets_data[comp_len][comp_pos]['constraints_keys'])                                                      
                        itemsets_data[comp_len][comp_pos]['children'].append(len(itemsets_list)-1)
                    comp_pos += 1
                cli -=1

    roots = []
    # leaves = []
    for pii, itemset in enumerate(itemsets_list):
        if itemset['parent'] is None:
            roots.append(pii)
            itemset['diff'] = list(itemset['constraints_keys'])
        else:
            itemset['parent'] = trans_pos[itemset['parent']]
        # if len(itemset['children']) == 0:
        #     leaves.append(pii)

    return roots, itemsets_list


#################### SELECTION OF FINAL CLAUSES
##########################################################
def selectClauses(clauses, graph, blue_path_covered, edges, parameters):
    covered_tuples = set()
    selected = []
    store_done_edges = {}

    remaining_to_cover = set(blue_path_covered)
    all_to_cover = set(blue_path_covered)
    pids = [(pid, parameters["try_simply"]!=0) for pid in range(len(clauses))]
    tmp_sel = []
    while len(pids) > 0 and ( parameters['limit_contrib'] < 0 or len(remaining_to_cover) > parameters['limit_contrib']):
        pids.sort(key=lambda x: clauses[x[0]].orderKey(parameters))
        pid, remap = pids.pop(0)
        curr_clause = clauses[pid]
        if not parameters.has_key('clause_oneback') \
               or parameters['clause_oneback'].support() != curr_clause.support() \
               or len(set(parameters['clause_oneback'].getTypesPred()) & set(curr_clause.getTypesPred())) == 0:
            bestP = remaining_to_cover & curr_clause.support()
            if parameters['limit_contrib'] < 0:
                tmp_contrib = max(1, abs(parameters['limit_contrib'])*len(all_to_cover & curr_clause.support()))
            else:
                tmp_contrib = parameters['limit_contrib']
            if len(bestP) >= tmp_contrib:

                if remap:
                    max_score = parameters['limit_abs_instances']
                    if parameters['limit_unit_instances'] != 0 and parameters['limit_unit_instances']*len(curr_clause.support()) < max_score:
                        max_score = parameters['limit_unit_instances']*len(curr_clause.support())

                    if parameters["try_simply"] < 0:
                        sim_map, details = graph.mapClause(curr_clause.constraints.getConstraintsList(), parameters['query_nodes'], None, True, max_score)
                        if details['status'] > 0:
                            pids.append((pid, False))
                            store_done_edges[pid] = details["done"]
                    else:
                        sid = 0
                        clauses[pid] = None
                        pvc = None
                        details = {'status': -1}
                        for sim_map, details in graph.generateClauses(curr_clause.constraints.getConstraintsList(), parameters['query_nodes'], True, max_score):
                            if details['status'] > 0:
                                sid +=1
                                pnodes_map = sim_map.projectFragment(details['head_fragment'])
                                pconstraints = Graph()
                                
                                for elem, atts in details['constraints'].items():
                                    if ( Graph.isEG(elem) and elem[0] in pnodes_map.nodesSet() and elem[1] in pnodes_map.nodesSet() ) or \
                                           ( not Graph.isEG(elem) and elem in pnodes_map.nodesSet()):
                                        for att in atts.items():
                                            pconstraints.parseElem(Graph.typeE(elem), elem, att[0], att[1])

                                clause_data = Clause("%sr%d" % (curr_clause.getId(), sid), pconstraints, pnodes_map, parameters['query_nodes'], edges)

                                accept, log_check = clause_data.checkAccept(parameters)
                                parameters["log"].printL(2, "%s %s Accept:%s" % ( "\t", clause_data.dispShort(), accept))
                                parameters["log"].printL(3, "%s [detail:%s]" % ("\t" , log_check))
                                if accept:
                                    pids.append((len(clauses), False))
                                    store_done_edges[len(clauses)] = list(details["done"])
                                    clauses.append(clause_data)

                        parameters["log"].printL(4, "Found %d alternatives for %s..." % (sid, curr_clause.dispShort()))

                    if details['status'] <= 0:
                        parameters["log"].printL(3, "Could not remap %s..." %  curr_clause.dispShort())
                    
                else:
                    remaining_to_cover -= curr_clause.support()
                    tmp_sel.append(pid)

    # inclus_patt = set()
    # for posA, pidA in enumerate(tmp_sel):
    #     for posB, pidB in enumerate(tmp_sel[:posA]):
    #         if clauses[pidB].blueTuples() <= clauses[pidA].blueTuples():
    #             if clauses[pidB].nbRedTuples() > 0 or clauses[pidA].nbRedTuples() == 0:
    #                 inclus_patt.add(pidB)
    #         elif clauses[pidA].blueTuples() <= clauses[pidB].blueTuples():
    #             if clauses[pidA].nbRedTuples() > 0 or clauses[pidB].nbRedTuples() == 0:  
    #                 inclus_patt.add(pidA)
                
    # for ni, patt in enumerate(sorted(set(tmp_sel) - inclus_patt, key=lambda x: clauses[x].orderKey(parameters))):
    for ni, patt in enumerate(tmp_sel):
        if parameters['limit_clause'] < 1 or ni < parameters['limit_clause']:

            #### CHECK FOR HANGING NODES
            hg = clauses[patt].constraints.hanging(parameters['query_nodes'])
            if len(hg) > 0:
                done_edges = store_done_edges[patt]
                constn = [(n,a) for n,a in clauses[patt].constraints.elems[Graph.ND].items() if n not in hg and len(a) > 0] 
                constn += [(e, clauses[patt].constraints.elems[Graph.EG][e]) for e in done_edges if (e[0] not in hg and e[1] not in hg) and e in clauses[patt].constraints.elems[Graph.EG]]
            #############
                pconstraints = Graph()                                
                sim_map, details = graph.mapClauseStraight(constn, parameters['query_nodes'])
                clause_data = Clause("%sh" % clauses[patt].getId(), pconstraints, sim_map, parameters['query_nodes'], edges)
                if len(clause_data.support()) <= len(clauses[patt].support()):
                    for elem, atts in constn:
                        for att in atts.items():
                            clause_data.constraints.parseElem(Graph.typeE(elem), elem, att[0], att[1])
                    parameters["log"].printL(5, "### CUT hanging in %s..." %  patt)
                    clauses[patt] = clause_data

            selected.append(patt)
            covered_tuples |= clauses[patt].support()
    return selected, covered_tuples


########################################################
################# WRITING
def writeClausesList(clauses, filename):
    fh = open(filename, "w")
    pids = [i for i, c in enumerate(clauses) if c is not None]
    pids.sort(key=lambda x: clauses[x].orderKey())
    for pid in pids:
        fh.write("\n".join(sorted(clauses[pid].listBMG(r'^([a-z])([0-9A-Z:.]*)$', 'T_' + clauses[pid].getId() +":"+r'\1\2'))) + "\n")
    fh.close()

def writePropPaths(paths, supps, fp, fc):
    for pi, path in enumerate(paths):
        fp.write("\n".join(sorted(path[0].listBMG(r'^([a-z])([0-9A-Z:.]*)$', 'T_' + str(pi) +":"+r'\1\2'))) + "\n")
        fc.write(" ".join(map(str,sorted(path[1]))) + "\n")
    fp.write("\n#### EDGES")
    for e, ei in supps.items():
        fp.write("\n# %d %s" % (ei, e))
        
def writeShortResults(edges, details, parameters, fh, fs=None, offset=0, coords=True):
    verti_step = 1000
    hori_step = 0
    with_infos = [t[0] for t in parameters.get('list_okey', [])+parameters.get('list_accept', [])]

    fh.write("#### SUMMARY\tEdge:%s Support:%f Blue:%d/%d/%d Red:%d Clauses:%d Explored:%d/%d Time:%s/%s/%s\n" % (parameters['id_exp'], parameters['thres_support'], len(set(details['edges']) & details['covered_tuples']), len(details['path_covered_blues']), len(details['edges']), len(details['covered_tuples'] - set(details['edges'])), len(details['selected']), parameters['count_ex'], parameters['count_cl'], (details['toc']-details['tic']), (details['tac']-details['toc']), (details['tac']-details['tic'])))

    for pid in details['selected']:
        clause  = details['clauses'][pid]

        ### SUPPORT OUTPUT
        if fs is not None:
            fs.write('%s\t' % clause.getId())
            for pair in sorted(clause.blueTuples()):
                fs.write('%s,%s ' % (pair[0], pair[1]))
            fs.write('\t')
            for pair in sorted(clause.redTuples()):
                fs.write('%s,%s ' % (pair[0], pair[1]))
            fs.write('\t')
            for pair in sorted(clause.outTuples()):
                fs.write('%s,%s ' % (pair[0], pair[1]))
            fs.write('\n')
            fs.flush()
            
        ### GRAPH OUTPUT
        if coords:
            fh.write("# _attributes T_%s pos=%d,%d\n" % (clause.getId(), (hori_step*offset)-700, (verti_step*offset)))
            fh.write("# _attributes T_%d:%s pos=%d,%d\n" % (offset, parameters['query_nodes'][0], (hori_step*offset)-500, (verti_step*offset)))
            fh.write("# _attributes T_%d:%s pos=%d,%d\n" % (offset, parameters['query_nodes'][-1], (hori_step*offset)+500, (verti_step*offset)))
        fh.write("\n".join(sorted(clause.listBMG(r'^([a-z])([0-9A-Z:.]*)$', ('T_%d:' % offset) +r'\1\2', None, with_infos))) + "\n")
        offset += 1
    fh.flush()
    return offset
